



python -c "import discretize; print(discretize.__version__)"
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "from importlib.metadata import version; print(version('discretize'))"
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
